from scapy.layers.inet import IP, TCP
from scapy.layers.l2 import Ether, ARP
from scapy.all import sniff
from collections import defaultdict
import duckdb as duckdb
import database_connection as dbConn
import time

connection = duckdb.connect('packets_data.duckdb')

reserved_ips = ["192.168.1.4", "192.168.1.1", "192.168.1.7", "172.16.0.3"]

syn_counts = defaultdict(int)
syn_counts_last_checked = time.time()

arp_table = defaultdict()

def ip_spoofing(src_mac, src_ip, packet):
	if src_ip not in reserved_ips:
		if src_ip.startswith("10.") or src_ip.startswith("192.168.") or src_ip.startswith("169.254."):
			dbConn.update_address(connection=connection, packet=packet, is_dangerous=True, type_of_attack="IP Spoofing")
			print("Possible IP spoofing using private networks detected.")
		elif src_ip.startswith("172."):
			octet = int(src_ip.split(".")[1])
			if 16 <= octet <= 31:
				dbConn.update_address(connection=connection, packet=packet, is_dangerous=True, type_of_attack="IP Spoofing")
				print("Possible IP spoofing using private networks detected.")

def syn_fin(packet):
	if TCP in packet:
		tcp_flags = packet[TCP].flags
		if 'S' in tcp_flags and 'F' in tcp_flags:
			dbConn.update_address(connection=connection, packet=packet, is_dangerous=True, type_of_attack="SYN-FIN")
			print("Malicious packet detected: SYN-FIN combination.")
			print(f"MAC Address of malicious agent: {packet[Ether].src}")

def null_packet(packet):
	if TCP in packet:
		tcp_flags = packet[TCP].flags
		if tcp_flags == 0:
			dbConn.update_address(connection=connection, packet=packet, is_dangerous=True, type_of_attack="Null Packet")
			print("Malicious null packet found.")
			print(f"MAC Address of malicious agent: {packet[Ether].src}")

def port_check(packet):
	if TCP in packet:
		src_port = packet[TCP].sport
		dst_port = packet[TCP].dport
		if src_port == 0 or dst_port == 0:
			dbConn.update_address(connection=connection, packet=packet, is_dangerous=True, type_of_attack="TCP Port 0")
			print("Illegal packet with source or destination port 0.")
			print(f"MAC Address of malicious agent: {packet[Ether].src}")

def destination_check(packet):
	if IP in packet:
		dest_ip = packet[IP].dst
		if dest_ip.endswith(".0") or dest_ip.endswith(".255"):
			dbConn.update_address(connection=connection, packet=packet, is_dangerous=True, type_of_attack="Broadcast Address")
			print("Packets with broadcast destination address detected.")
			print(f"MAC Address of malicious agent: {packet[Ether].src}")

# Simple Check to see if the Packet has a valid format
def malformed_check(packet):
	try:
		# Attempt to parse packet
		packet.show()
	except Exception as e:
		dbConn.update_address(connection=connection, packet=packet, is_dangerous=True, type_of_attack="Malformed Packet")
		print(f"Malformed packet detected: {e}")
		print(f"Source: {packet[IP].src if IP in packet else 'Unknown'}")

# Detect if a certain IP is sending a suspiciously large amount of SYN packets
def syn_flood_detection(packet):
	global syn_counts_last_checked
	if TCP in packet and packet[TCP].flags == 'S':
		syn_counts[packet[IP].src] += 1

	# Check counts periodically (every 10s)
	if time.time() - syn_counts_last_checked > 10:
		for ip, count in syn_counts.items():
			if count > 100:
				dbConn.update_address(connection=connection, packet=packet, is_dangerous=True, type_of_attack="SYN Flood")
				print(f"Possible SYN flood from {ip}. Number of SYN Packets in last {time.time() - syn_counts_last_checked} seconds: {count}")
		syn_counts.clear()
		syn_counts_last_checked = time.time()

# Checks if the same IP has a consistent MAC.
# Otherwise, an attacker might try to use an IP already in use, like e.g. the gateway's IP, to intercept traffic
def arp_spoofing(packet):
	# op == 2 checks for an ARP reply (they say what their MAC is)
	if ARP in packet and packet[ARP].op == 2:
		src_ip = packet[ARP].psrc
		src_mac = packet[ARP].hwsrc
		if src_ip in arp_table:
			if arp_table[src_ip] != src_mac:
				dbConn.update_address(connection=connection, packet=packet, is_dangerous=True, type_of_attack="ARP Spoofing")
				print(f"Possible ARP spoofing detected from IP {src_ip}. Conflicting MACs: {arp_table[src_ip]} and {src_mac}")
		else:
			arp_table[src_ip] = src_mac

def tcp_reset_attack(packet):
	if packet[TCP].flags == 'R':
		dbConn.update_address(connection=connection, packet=packet, is_dangerous=True, type_of_attack="TCP Reset")
		print(f"Possible TCP reset attack from {packet[IP].src}")

def packet_handler(packet):
	dbConn.add_address(connection=connection, packet=packet)
	if IP in packet:
		src_ip = packet[IP].src
		dst_ip = packet[IP].dst
		src_mac = packet[Ether].src if Ether in packet else None
		print(f"Packet captured: {src_ip} -> {dst_ip}")
		ip_spoofing(src_mac, src_ip, packet)
		destination_check(packet)
		malformed_check(packet)
		syn_flood_detection(packet)
		arp_spoofing(packet)
	if TCP in packet:
		syn_fin(packet)
		null_packet(packet)
		port_check(packet)
		tcp_reset_attack(packet)

# Allows unit testing of the dict
def get_dict():
	return syn_counts
def dict_clear():
	syn_counts.clear()

def main():
	dbConn.initiate_database(connection)
	print("Starting packet capture...")
	sniff(prn=packet_handler, store=False)

if __name__ == "__main__":
	main()
