import time
import unittest

from scapy.layers.inet import IP, TCP
from unittest.mock import patch
from scapy.layers.l2 import Ether

from packet_capturing import packet_handler, get_dict, dict_clear, syn_flood_detection, port_check, ip_spoofing, \
    syn_fin, null_packet, tcp_reset_attack

class TestPacketCapturing(unittest.TestCase):

    # This method is called before every test
    def setUp(self):
        dict_clear()

    def test_baseCase(self):
        # Create a dummy packet
        # Ether addresses are placeholders
        packet = (IP(src = "100.84.6.141", dst = "192.168.1.1") /
                  TCP(dport = 80, sport = 12345, flags = "S") /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff"))
        # Show detailed information about the package
        print(packet.show())

        # Mockito checks if print was called
        with patch('builtins.print') as mock_print:
            packet_handler(packet)
            mock_print.assert_any_call("Packet captured: 100.84.6.141 -> 192.168.1.1")

    def test_ip_spoofing(self):
        # Starts with 169.254, it's a suspicious packet
        packet = IP(src = "169.254.6.141", dst = "192.168.1.1")

        with patch('builtins.print') as mock_print:
            ip_spoofing(0, packet[IP].src)
            mock_print.assert_any_call("Possible IP spoofing using private networks detected.")

    def test_syn_fin(self):
        # Has S and F flags, which is also known as SYN-FIN, which is a malicious packet signature

        packet = (IP(src = "100.84.6.141", dst = "192.168.1.1") /
                  TCP(dport = 80, flags = "SF") /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff"))

        with patch('builtins.print') as mock_print:
            syn_fin(packet)
            mock_print.assert_any_call("Malicious packet detected: SYN-FIN combination.")

    def test_null_packet(self):
        packet = (IP(src = "100.84.6.141", dst = "192.168.1.1") /
                  TCP(dport = 80, flags = 0) /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff"))

        with patch('builtins.print') as mock_print:
            null_packet(packet)
            mock_print.assert_any_call("Malicious null packet found.")

    def test_port_dest_check(self):
        packet = (IP(src="100.84.6.141", dst="192.168.1.1") /
                  TCP(dport = 0, sport = 200, flags="F") /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff"))

        with patch('builtins.print') as mock_print:
            port_check(packet)
            mock_print.assert_any_call("Illegal packet with source or destination port 0.")

    def test_port_src_check(self):
        packet = (IP(src="100.84.6.141", dst="192.168.1.1") /
                  TCP(dport=80, sport=0, flags="F") /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff"))

        with patch('builtins.print') as mock_print:
            packet_handler(packet)
            mock_print.assert_any_call("Illegal packet with source or destination port 0.")

    def test_syn_flood_detection(self):
        repetitions = 150
        packet = (IP(src="100.84.6.141", dst="192.168.1.1") /
                  TCP(dport=80, sport=123, flags="S") /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff"))


        with patch('builtins.print') as mock_print:
            for i in range(repetitions):
                syn_flood_detection(packet)
            time.sleep(10)
            syn_flood_detection(packet)
            mock_print.assert_called()

        self.assertEqual({}, get_dict(), "Expected the packet's IP")

    def test_tcp_reset_attack(self):
        packet = (IP(src="100.84.6.141", dst="192.168.1.1") /
                  TCP(dport=80, sport=0, flags="R") /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff"))

        with patch('builtins.print') as mock_print:
            tcp_reset_attack(packet)
            mock_print.assert_any_call(f"Possible TCP reset attack from {packet[IP].src}")

if __name__ == '__main__':
    unittest.main()
